# -*- coding: utf-8 -*-

"""
    CyInstaller.log
    ~~~~~~~~~~~~~~~~~~~
    Log of CyInstaller
"""

import logging
import sys
from logging import Logger
from typing import Callable


class CyInstallerLogger(Logger):
    def loglist(self, msg: str, lists: list, itemfmt: Callable[[object], str] = None,
                level=logging.INFO) -> None:
        details = list(' - %s' % (itemfmt(item) if itemfmt is not None else item) for item in lists)
        self.log(level, '%s\n%s' % (msg, '\n'.join(details)))

    def write(self, output_stream):
        self.log(self.level, output_stream)

    def flush(self):
        pass


FORMAT = '%(asctime)s %(levelname)s: %(message)s'
DATEFMT = '%Y-%m-%d %H:%M:%S'
TRACE = logging.TRACE = logging.DEBUG - 5

logging.addLevelName(TRACE, 'TRACE')
logging.setLoggerClass(CyInstallerLogger)
logging.basicConfig(format=FORMAT, level=logging.INFO, datefmt=DATEFMT, stream=sys.stdout)


# noinspection PyPep8Naming, PyTypeChecker
def getLogger(name: str) -> CyInstallerLogger:
    from logging import getLogger as baseGetLogger
    return baseGetLogger(name)



# -*- coding: utf-8 -*-

"""
    cyinstaller.core
    ~~~~~~~~~~~~~~~~~~~
     The core methods of CyInstaller
"""

import json
import re
import os
import platform
import sys

from pathlib import Path
from typing import Union
from shutil import copyfile, rmtree, copytree
from PyInstaller.compat import PY3_BASE_MODULES

from .log import getLogger

_suffix = '.pyd' if platform == 'win32' else '.so'
_dotpath = Path('')

PY_BASE_MODULES = PY3_BASE_MODULES
logger = getLogger(__name__)


def glob_copy(source: Path, target: Path, rglob: str = '*') -> None:
    if source.is_file():
        os.makedirs(target.parent, exist_ok=True)
        copyfile(source, target)
        return

    for x in source.glob(rglob):
        dist = target.joinpath(str(x)[len(str(source)) + 1:], '')
        os.makedirs(dist.parent, exist_ok=True)
        if x.is_file():
            copyfile(x, dist)
        else:
            copytree(str(x), str(dist), dirs_exist_ok=True)


def add_hidden_imports(imports: Union[str, list], hiddenimports: list) -> None:
    if isinstance(imports, str) and str not in hiddenimports and str not in PY_BASE_MODULES:
        hiddenimports.append(imports)
    elif isinstance(imports, list):
        for imp in imports:
            add_hidden_imports(imp, hiddenimports)


def parse_hidden_imports(sources: [str]) -> [str]:
    imports = []

    regex = re.compile(r"^\s*(?:from|import)\s+(\S+(?:\s*,\s*\w+)*)")
    for x in sources:
        with open(x) as src:
            for line in src.readlines():
                match = regex.search(line)
                if match and len(match.groups()) > 0:
                    if not match.group(1) in imports and not match.group(1).startswith('.'):
                        imports.append(match.group(1))

    return imports


def find_packages(base: Path, sources: [str]) -> [str]:
    pkgs = []
    scaned = []

    def _detect_package(source: Union[str, Path]) -> str:
        m = Path(source).parent.parts
        for p in m:
            if p in pkgs:
                return '.'.join(m[m.index(p):])
        return '.'.join(m)

    for src in sources:
        src = Path(src)

        if src.exists() and src.name == '__init__.py':
            if src.parent not in scaned:
                pkgs.append(_detect_package(str(src)))
                scaned.append(src.parent)
            continue

        for item in base.glob(str(src)):
            if item.parent in scaned:
                continue

            init_file = item.parent.joinpath('__init__.py')
            if init_file.exists():
                pkgs.append(_detect_package(init_file))
            scaned.append(item.parent)

    pures = []
    for item in pkgs:
        pure = []
        part = Path(os.sep.join(item.split('.')))
        while part.stem != '':
            if not part.joinpath('__init__.py').exists():
                break
            pure.insert(0, part.stem)
            part = part.parent
        if len(pure) > 0:
            pures.append('.'.join(pure))

    return pures


def build(yaml_file='setup.yml'):
    root_dir = os.getcwd()
    sys.stdout = logger

    # read config
    logger.info('[STAGE 01] Read config from yaml file - [%s].', yaml_file)

    from .config import Config
    config = Config(yaml_file)

    # set build dir
    logger.info('[STAGE 02] Prepare building dir - [%s].', config.stage_path)

    build_path = config.stage_path
    if build_path.exists():
        rmtree(build_path)
    build_path.mkdir()

    # set dist dir
    dist_path = Path(root_dir).joinpath(config.dist)
    if dist_path.exists():
        rmtree(dist_path)

    # run cython build
    c_options = {
        'sources': config.compiles.raw_includes,
        'exclude': config.compiles.raw_excludes,
        'build_dir': str(config.stage_cython_temp_path),
        **config.stage_cython_options
    }

    logger.info('[STAGE 03] Execute cython compile with following options: \n %s', json.dumps(c_options, indent='  '))

    from distutils.core import setup
    from Cython.Build import cythonize

    ext_module = cythonize(config.compiles.raw_includes,
                           exclude=config.compiles.raw_excludes, build_dir=str(config.stage_cython_temp_path),
                           **config.stage_cython_options)

    script_args = ['build_ext', '-t', str('.'), '-b', str(config.stage_cython_path)]
    setup(name=config.name, ext_modules=ext_module, script_args=script_args)

    if not config.stage_cython_path.exists():
        raise BrokenPipeError('Unable build source code with cython.')

    if not config.stage_debug:
        rmtree(config.stage_cython_temp_path)

    # prepare soures for PyInstaller
    p_sources = {}

    for item in (config.packages.pures + [config.entrypoint]):
        target = config.stage_cython_path.joinpath(item.to.joinpath(item.source))
        p_sources.update({item.realsource: item.to.joinpath(item.source)})
        glob_copy(item.realsource, target)

    p_sources.pop(config.entrypoint.realsource)  # just copy entrypoint

    logger.loglist('[STAGE 04] Prepare package sources for PyInstaller ...',
                   list('%s->%s' % (k, v) for k, v in p_sources.items()))

    # prepare binaries for PyInstaller
    p_binaries = {}

    for item in config.binaries.pures:
        target = config.stage_cython_path.joinpath(item.to.joinpath(item.source))
        p_binaries.update({item.realsource: item.to.joinpath(item.source)})
        glob_copy(item.realsource, target)

    if config.cython_binaries:
        for item in config.stage_cython_path.rglob('*' + _suffix):
            f = item.relative_to(config.stage_cython_path)
            p_binaries.update({f: f})

    logger.loglist('[STAGE 05] Prepare binaries for PyInstaller ...',
                   list('%s->%s' % (k, v) for k, v in p_binaries.items()))

    # prepare datas for PyInstaller
    p_datas = {}

    for item in config.datas.pures:
        target = config.stage_cython_path.joinpath(item.to.joinpath(item.source))
        p_datas.update({item.realsource: item.to.joinpath(item.source)})
        glob_copy(item.realsource, target)

    logger.loglist('[STAGE 06] Prepare datas for PyInstaller ...', list('%s->%s' % (k, v) for k, v in p_datas.items()))

    # prepare hiddenimports for PyInstaller
    p_hiddenimports = config.hiddenimports
    if config.auto_hiddenimports:
        srcfiles = list(x.realsource for x in (config.compiles.pures + config.packages.pures))
        add_hidden_imports(parse_hidden_imports(srcfiles), p_hiddenimports)

    _app_packages = find_packages(config.root, [x.realsource for x in config.compiles.pures])
    add_hidden_imports(_app_packages, p_hiddenimports)

    logger.loglist('[STAGE 07] Prepare hiddenimports for PyInstaller ...', p_hiddenimports)

    # prepare PyInstaller spec file
    p_template = config.stage_pyinstaller_template
    if p_template is None:
        p_template = Path(os.path.dirname(__file__)).joinpath('template.spec')
    else:
        p_template = Path(p_template)
    copyfile(p_template, config.stage_cython_path.joinpath('setup.spec'))

    os.chdir(config.stage_cython_path)

    with open('setup.spec', 'r+') as template:
        content = template.read()
        content = re.sub(r'\${app_name}', '"%s"' % config.name, content)
        content = re.sub(r'\${sources}', str([config.entrypoint.source]), content)
        content = re.sub(r'\${hiddenimports}', str(p_hiddenimports), content)
        content = re.sub(r'\${datas}',
                         str([(str(x), str(x.parent if x.is_file() else x)) for x in p_datas.values()]), content)
        content = re.sub(r'\${binaries}',
                         str([(str(x), str(x.parent if x.is_file() else x)) for x in p_binaries.values()]), content)
        template.seek(0)
        template.write(content)
        template.truncate()

    logger.info('[STAGE 08] Generate template spec file for PyInstaller: [%s]',
                config.stage_cython_path.joinpath('setup.sepc'))

    # run PyInstaller with built arguments
    logger.info('[STAGE 09] Execute PyInstaller, detail options see the spec file.')

    from PyInstaller.building.build_main import main

    main(None, 'setup.spec', **{
        'distpath': str(dist_path),
        'workpath': str(Path(root_dir).joinpath(config.stage_pyinstaller_path)),
        'noconfirm': False,
    })

    # copy other distribution files to target dir
    os.chdir(root_dir)
    relates = {}

    for item in config.relates.pures:
        target = config.dist.joinpath(item.to.joinpath(item.source))
        relates.update({item.realsource: item.to.joinpath(item.source)})
        glob_copy(item.realsource, target)

    logger.loglist('[STAGE 10] Copy relates to dist...', list('%s->%s' % (k, v) for k, v in relates.items()))

    # clean up build dir
    if not config.stage_debug:
        rmtree(config.stage_cython_path)
        rmtree(config.stage_pyinstaller_path)

    logger.info('[FINISHED] CyInstaller execute finished, distribution files are in - [%s]', config.dist.absolute())

    # restore stdout
    sys.stdout = sys.__stdout__



# -*- coding: utf-8 -*-

"""
    cyinstaller.config
    ~~~~~~~~~~~~~~~~~~~
     The custom config loader of CyInstaller
"""

import os
from pathlib import Path
from typing import Union

import yaml

_default_config = {
    'app': 'CyInstaller Application',
    'root': '.',
    'compiles': [],
    'packages': [],
    'datas': [],
    'binaries': [],
    'cython_binaries': True,
    'relates': [],
    'hiddenimports': [],
    'modules': [],
    'auto_hiddenimports': True,

    'stage': {
        'path': '_build',
        'debug': True,
        'cython': {
            'path_tmp': 'compile',
            'path': 'cython',
            'options': {
                'compiler_directives': {
                    'always_allow_keywords': True
                },
                'nthreads': 0,
                'language_level': 3,
            }
        },
        'pyinstaller': {
            'path': 'package',
            'template': None,
        },
        'dist': 'target'
    }
}


def merge_dict(src: dict, default: dict):
    dist = src.copy()

    for k, v in dist.items():
        if v is None:
            dist[k] = default.get(k, None)
            continue

        t = default.get(k, None)
        if t is None:
            continue

        if isinstance(v, dict):
            dist[k] = merge_dict(v, t)
        if isinstance(v, (tuple, list)):
            if isinstance(t, (tuple, list)):
                v.extend(t)
            else:
                v.append(v)
            dist[k] = v
        elif type(v) is not type(t):
            raise ValueError('Config %s value type error.' % k)

    for k, v in default.items():
        if k not in dist.keys():
            dist[k] = v

    return dist


class PortableSource:
    def __init__(self, src: str, base=Path(''), to=Path('')) -> None:
        self._exclude = src.startswith('!')

        if self._exclude:
            src = src[1:]

        _mapping = src.split('::')

        self._port = base if len(_mapping) == 1 else base.joinpath(_mapping[0])

        _source = (_mapping[0] if len(_mapping) == 1 else _mapping[1]).split('->')

        self._source = _source[0]
        self._to = to if len(_source) == 1 else to.joinpath(_source[1])
        self._realsource = self._port.joinpath(self._source)

    @property
    def origin(self):
        if self._to is None:
            return '%s::%s' % (self._port, self._source)
        return '%s::%s->%s' % (self._port, self._source, self._to)

    @property
    def realsource(self) -> Path:
        return self._realsource

    @property
    def port(self) -> Path:
        return self._port

    @property
    def source(self) -> str:
        return self._source

    @property
    def to(self) -> Path:
        return self._to

    @property
    def is_exclude(self):
        return self._exclude

    def __eq__(self, other):
        return self.origin == other.origin

    def __hash__(self):
        return self.origin.__hash__()

    def __str__(self):
        return self.origin

    def __repr__(self):
        return self.origin


class SourceList:
    def __init__(self, sources: [Union[str, PortableSource]]) -> None:
        item = sources[0] if len(sources) > 0 else None

        if isinstance(item, str):
            s = [PortableSource(x) for x in sources]
        else:
            s = sources

        self._list = list(s)

    def extend(self, others):
        self._list.extend(others.all)

    @property
    def all(self) -> [PortableSource]:
        return self._list

    @property
    def includes(self) -> [PortableSource]:
        return [x for x in self._list if not x.is_exclude]

    @property
    def raw_includes(self) -> [str]:
        return [str(x.realsource) for x in self.includes]

    @property
    def excludes(self) -> [PortableSource]:
        return [x for x in self._list if x.is_exclude]

    @property
    def raw_excludes(self) -> [str]:
        return [str(x.realsource) for x in self.excludes]

    @property
    def pures(self) -> [PortableSource]:
        includes = list()
        for inc in self.includes:
            for src in inc.port.glob(inc.source):
                includes.append(PortableSource('%s::%s->%s' % (inc.port, src.relative_to(inc.port), inc.to)))
        excludes = list()
        for exc in self.excludes:
            for src in exc.port.glob(exc.source):
                includes.append(PortableSource('%s::%s->%s' % (exc.port, src.relative_to(exc.port), exc.to)))

        return [item for item in includes if item not in excludes]

    def mappings(self, to=None) -> [PortableSource]:
        includes = list()
        for inc in self.includes:
            to = inc.to if to is None else to
            if isinstance(to, str):
                to = Path(to)

            for src in inc.port.glob(inc.source):
                includes.append(PortableSource('%s::%s->%s' % (inc.port, src.relative_to(inc.port), to)))

        excludes = list()
        for exc in self.excludes:
            to = exc.to if to is None else to
            if isinstance(to, str):
                to = Path(to)

            for src in exc.port.glob(exc.source):
                includes.append(PortableSource('%s::%s->%s' % (exc.port, src.relative_to(exc.port), to)))

        return [item for item in includes if item not in excludes]

    def __str__(self):
        return str([x for x in self.all])


class ModuleConfig:
    def __init__(self, config: dict) -> None:
        self._base = Path(config.get('base', None))
        self._package = config.get('package', None)
        self._package_from_base = config.get('package_from_base', True)

        if self._base is None:
            raise SyntaxError('The base path of module is None.')

        if self._package is None:
            raise SyntaxError('The package name of module in % is None.' % self._base)

        if not self._package_from_base:
            self._packagebase = self._base.joinpath(self._package)

        self._compiles = SourceList(
            [PortableSource(x, self._packagebase, self._package) for x in config.get('compiles') or []])
        self._packages = SourceList(
            [PortableSource(x, self._base, self._package) for x in config.get('packages') or []])
        self._binaries = SourceList(
            [PortableSource(x, self._base, self._package) for x in config.get('binaries') or []])
        self._datas = SourceList(
            [PortableSource(x, self._base, self._package) for x in config.get('datas') or []])
        self._relates = SourceList(
            [PortableSource(x, self._base, self._package) for x in config.get('relates') or []])

    @property
    def base(self) -> Path:
        return self._base

    @property
    def package(self) -> str:
        return self._package

    @property
    def package_from_base(self):
        return self._package_from_base

    @property
    def compiles(self) -> SourceList:
        return self._compiles

    @property
    def packages(self) -> SourceList:
        return self._packages

    @property
    def binaries(self) -> SourceList:
        return self._binaries

    @property
    def datas(self) -> SourceList:
        return self._datas

    @property
    def relates(self) -> SourceList:
        return self._relates


class Config:
    def __init__(self, yaml_file='setup.yml') -> None:
        if not os.path.exists(yaml_file):
            raise EnvironmentError('CyInstaller yaml file not exists.')

        try:
            with open(Path(os.getcwd()).joinpath(yaml_file), 'r') as cfg:
                config = yaml.safe_load(cfg)
                self._config = merge_dict(config.get('setup', {}), _default_config)

                self._modules = []
                for m in self._config['modules']:
                    self._modules.append(ModuleConfig(m))

                if self._config.get('entrypoint', None) is None:
                    raise ValueError('The entrypoint not set.')

                self._entrypoint = PortableSource(self._config['entrypoint'])

        except Exception as e:
            raise SyntaxError('Error load CyInstaller configuration from yaml - %s' % e)

    @property
    def name(self) -> str:
        return self._config['name']

    @property
    def root(self) -> Path:
        return Path(self._config['root'])

    @property
    def compiles(self) -> SourceList:
        collection = SourceList([])
        for module in self._modules:
            collection.extend(module.compiles)
        collection.extend(SourceList([PortableSource(x) for x in self._config['compiles']]))
        return collection

    @property
    def packages(self) -> SourceList:
        collection = SourceList([])
        for module in self._modules:
            collection.extend(module.packages)
        collection.extend(SourceList([PortableSource(x) for x in self._config['packages']]))
        return collection

    @property
    def binaries(self) -> SourceList:
        collection = SourceList([])
        for module in self._modules:
            collection.extend(module.binaries)
        collection.extend(SourceList([PortableSource(x) for x in self._config['binaries']]))
        return collection

    @property
    def cython_binaries(self) -> bool:
        return self._config['cython_binaries']

    @property
    def datas(self) -> SourceList:
        collection = SourceList([])
        for module in self._modules:
            collection.extend(module.datas)
        collection.extend(SourceList([PortableSource(x) for x in self._config['datas']]))
        return collection

    @property
    def relates(self) -> SourceList:
        collection = SourceList([])
        for module in self._modules:
            collection.extend(module.relates)
        collection.extend(SourceList([PortableSource(x) for x in self._config['relates']]))
        return collection

    @property
    def hiddenimports(self) -> [str]:
        return self._config['hiddenimports']

    @property
    def auto_hiddenimports(self) -> bool:
        return self._config['auto_hiddenimports']

    @property
    def modules(self) -> [ModuleConfig]:
        return self._modules

    @property
    def entrypoint(self) -> PortableSource:
        return self._entrypoint

    @property
    def stage_path(self) -> Path:
        return Path(self._config['stage']['path'])

    @property
    def stage_debug(self) -> bool:
        return self._config['stage']['debug']

    @property
    def stage_cython_path(self) -> Path:
        return self.stage_path.joinpath(self._config['stage']['cython']['path'])

    @property
    def stage_cython_temp_path(self) -> Path:
        return self.stage_path.joinpath(self._config['stage']['cython']['path_tmp'])

    @property
    def stage_cython_options(self) -> dict:
        return self._config['stage']['cython']['options']

    @property
    def stage_pyinstaller_path(self) -> Path:
        return self.stage_path.joinpath(self._config['stage']['pyinstaller']['path'])

    @property
    def stage_pyinstaller_template(self) -> str:
        return self._config['stage']['pyinstaller']['template']

    @property
    def dist(self) -> Path:
        return self.root.joinpath(self._config['dist'])



# -*- coding: utf-8 -*-

"""
    CyInstaller.main
    ~~~~~~~~~~~~~~~~~~~
    CLI Entry point of CyInstaller
"""

import click
import sys


@click.command()
@click.option('--file', default='setup.yml', help='The setup config file, default is setup.yml')
def build_command(file: str):
    from CyInstaller import __version__
    from CyInstaller.log import getLogger

    logger = getLogger(__name__)

    try:
        logger.info('-------- Welcome to use CyInstaller v%s --------\n\n', __version__)

        from CyInstaller.core import build
        build(file)

    except (InterruptedError, KeyboardInterrupt):
        logger.info('CyInstaller executing interrupted.')
    except Exception as e:
        logger.info('CyInstaller executing interrupted with error - %s', e)

    # restore stdout
    print()
    sys.stdout = sys.__stdout__


if __name__ == "__main__":
    from pathlib import Path
    sys.path.append(str(Path('..').resolve()))
    build_command()
